##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
# Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
import unittest

from Acquisition import Implicit
from OFS.Folder import Folder


class DummyFolder(Folder):
    pass


class DummyPlugin(Implicit):
    pass


class PluginRegistryTests(unittest.TestCase):

    def _getTargetClass(self):

        from Products.PluginRegistry.PluginRegistry import PluginRegistry

        return PluginRegistry

    def _makeOne(self, plugin_info=None, *args, **kw):

        if plugin_info is None:
            plugin_info, IFoo, IBar = self._makePluginInfo()

        return self._getTargetClass()(plugin_info, *args, **kw)

    def _makePluginInfo(self):

        from zope.interface import Interface

        class IFoo(Interface):
            def foo():
                """ Foo. """

        class IBar(Interface):
            def bar():
                """ Bar. """

        _PLUGIN_INFO = ((IFoo, 'IFoo', 'foo', 'Foo test'),
                        (IBar, 'IBar', 'bar', 'Bar test'))

        return _PLUGIN_INFO, IFoo, IBar

    def test_class_conforms_to_IPluginRegistry(self):
        from zope.interface.verify import verifyClass

        from Products.PluginRegistry.interfaces import IPluginRegistry
        verifyClass(IPluginRegistry, self._getTargetClass())

    def test_instance_conforms_to_IPluginRegistry(self):
        from zope.interface.verify import verifyObject

        from Products.PluginRegistry.interfaces import IPluginRegistry
        verifyObject(IPluginRegistry, self._makeOne())

    def test_listPlugins_miss(self):
        from zope.interface import Interface

        class INonesuch(Interface):
            pass

        preg = self._makeOne(())
        self.assertRaises(KeyError, preg.listPlugins, INonesuch)

    def test_listPluginIds_miss(self):
        from zope.interface import Interface

        class INonesuch(Interface):
            pass

        preg = self._makeOne(())
        self.assertRaises(KeyError, preg.listPluginIds, INonesuch)

    def test_listPlugins_hit(self):
        plugin_info, IFoo, IBar = self._makePluginInfo()
        preg = self._makeOne(plugin_info)
        self.assertEqual(len(preg.listPluginIds(IFoo)), 0)

    def test_listPluginIds_hit(self):
        plugin_info, IFoo, IBar = self._makePluginInfo()
        preg = self._makeOne(plugin_info)
        self.assertEqual(len(preg.listPluginIds(IFoo)), 0)

    def test_listPluginTypeInfo_empty(self):
        pref = self._makeOne(())

        pti = pref.listPluginTypeInfo()
        self.assertEqual(len(pti), 0)

    def test_listPluginTypeInfo_filled(self):
        plugin_info, IFoo, IBar = self._makePluginInfo()
        pref = self._makeOne(plugin_info)

        pti = pref.listPluginTypeInfo()
        self.assertEqual(len(pti), 2)
        self.assertEqual(pti[0]['interface'], IFoo)
        self.assertEqual(pti[0]['title'], 'foo')
        self.assertEqual(pti[1]['interface'], IBar)
        self.assertEqual(pti[1]['title'], 'bar')

    def test_activatePlugin_no_child(self):
        plugin_info, IFoo, IBar = self._makePluginInfo()
        parent = DummyFolder()
        preg = self._makeOne(plugin_info).__of__(parent)

        self.assertRaises(AttributeError, preg.activatePlugin, IFoo,
                          'foo_plugin')

    def test_activatePluginInterface_non_conforming_interface(self):
        plugin_info, IFoo, IBar = self._makePluginInfo()
        parent = DummyFolder()
        foo_plugin = DummyPlugin()
        parent._setObject('foo_plugin', foo_plugin)
        preg = self._makeOne(plugin_info).__of__(parent)

        self.assertRaises(ValueError, preg.activatePlugin, IFoo, 'foo_plugin')

    def test_activatePlugin_valid_child(self):
        from zope.interface import directlyProvides
        plugin_info, IFoo, IBar = self._makePluginInfo()
        parent = DummyFolder()
        foo_plugin = DummyPlugin()
        directlyProvides(foo_plugin,  (IFoo,))
        parent._setObject('foo_plugin', foo_plugin)

        preg = self._makeOne(plugin_info).__of__(parent)

        preg.activatePlugin(IFoo, 'foo_plugin')

        idlist = preg.listPluginIds(IFoo)
        self.assertEqual(len(idlist), 1)
        self.assertEqual(idlist[0], 'foo_plugin')

        plugins = preg.listPlugins(IFoo)
        self.assertEqual(len(plugins), 1)
        plugin = plugins[0]
        self.assertEqual(plugin[0], 'foo_plugin')
        self.assertEqual(plugin[1], preg.foo_plugin)

    def test_activatePlugin_then_remove_interface(self):
        from zope.interface import directlyProvides
        plugin_info, IFoo, IBar = self._makePluginInfo()
        parent = DummyFolder()
        foo_plugin = DummyPlugin()
        directlyProvides(foo_plugin,  (IFoo,))
        parent._setObject('foo_plugin', foo_plugin)

        preg = self._makeOne(plugin_info).__of__(parent)

        preg.activatePlugin(IFoo, 'foo_plugin')

        replacement = DummyPlugin()
        parent._delObject('foo_plugin')
        parent._setObject('foo_plugin', replacement)

        idlist = preg.listPluginIds(IFoo)
        self.assertEqual(len(idlist), 1)  # note discrepancy

        plugins = preg.listPlugins(IFoo)
        self.assertEqual(len(plugins), 0)

    def test_deactivatePlugin(self):
        from zope.interface import directlyProvides
        plugin_info, IFoo, IBar = self._makePluginInfo()
        parent = DummyFolder()
        foo_plugin = DummyPlugin()
        directlyProvides(foo_plugin, (IFoo,))
        parent._setObject('foo_plugin', foo_plugin)

        bar_plugin = DummyPlugin()
        directlyProvides(bar_plugin, (IFoo,))
        parent._setObject('bar_plugin', bar_plugin)

        baz_plugin = DummyPlugin()
        directlyProvides(baz_plugin, (IFoo,))
        parent._setObject('baz_plugin', baz_plugin)

        preg = self._makeOne(plugin_info).__of__(parent)

        preg.activatePlugin(IFoo, 'foo_plugin')
        preg.activatePlugin(IFoo, 'bar_plugin')
        preg.activatePlugin(IFoo, 'baz_plugin')

        preg.deactivatePlugin(IFoo, 'bar_plugin')

        idlist = preg.listPluginIds(IFoo)
        self.assertEqual(len(idlist), 2)
        self.assertEqual(idlist[0], 'foo_plugin')
        self.assertEqual(idlist[1], 'baz_plugin')

    def test_movePluginsTop(self):
        from zope.interface import directlyProvides
        plugin_info, IFoo, IBar = self._makePluginInfo()
        parent = DummyFolder()
        foo_plugin = DummyPlugin()
        directlyProvides(foo_plugin, (IFoo,))
        parent._setObject('foo_plugin', foo_plugin)

        bar_plugin = DummyPlugin()
        directlyProvides(bar_plugin, (IFoo,))
        parent._setObject('bar_plugin', bar_plugin)

        baz_plugin = DummyPlugin()
        directlyProvides(baz_plugin, (IFoo,))
        parent._setObject('baz_plugin', baz_plugin)

        preg = self._makeOne(plugin_info).__of__(parent)

        preg.activatePlugin(IFoo, 'foo_plugin')
        preg.activatePlugin(IFoo, 'bar_plugin')
        preg.activatePlugin(IFoo, 'baz_plugin')

        self.assertRaises(ValueError,
                          preg.movePluginsTop, IFoo, ('quux_plugin',))

        # Moving one plugin to top has obvious result
        preg.movePluginsTop(IFoo, ('bar_plugin',))

        idlist = preg.listPluginIds(IFoo)
        self.assertEqual(len(idlist), 3)

        self.assertEqual(idlist[0], 'bar_plugin')
        self.assertEqual(idlist[1], 'foo_plugin')
        self.assertEqual(idlist[2], 'baz_plugin')

        # Moving more than one plugin to top puts them
        # one by one at the top
        # iow, last in the list gets to top
        preg.movePluginsTop(IFoo, ('bar_plugin', 'baz_plugin'))

        idlist = preg.listPluginIds(IFoo)
        self.assertEqual(len(idlist), 3)

        self.assertEqual(idlist[0], 'baz_plugin')
        self.assertEqual(idlist[1], 'bar_plugin')
        self.assertEqual(idlist[2], 'foo_plugin')

        # Moving the top plugin up should not change anything.
        preg.movePluginsTop(IFoo, ('baz_plugin',))
        idlist = preg.listPluginIds(IFoo)
        self.assertEqual(idlist,
                         ('baz_plugin', 'bar_plugin', 'foo_plugin'))

        # Moving the top plugin and another one could change something.
        preg.movePluginsTop(IFoo, ('baz_plugin', 'foo_plugin'))
        idlist = preg.listPluginIds(IFoo)
        self.assertEqual(idlist,
                         ('foo_plugin', 'baz_plugin', 'bar_plugin'))

    def test_movePluginsUp(self):
        from zope.interface import directlyProvides
        plugin_info, IFoo, IBar = self._makePluginInfo()
        parent = DummyFolder()
        foo_plugin = DummyPlugin()
        directlyProvides(foo_plugin, (IFoo,))
        parent._setObject('foo_plugin', foo_plugin)

        bar_plugin = DummyPlugin()
        directlyProvides(bar_plugin, (IFoo,))
        parent._setObject('bar_plugin', bar_plugin)

        baz_plugin = DummyPlugin()
        directlyProvides(baz_plugin, (IFoo,))
        parent._setObject('baz_plugin', baz_plugin)

        preg = self._makeOne(plugin_info).__of__(parent)

        preg.activatePlugin(IFoo, 'foo_plugin')
        preg.activatePlugin(IFoo, 'bar_plugin')
        preg.activatePlugin(IFoo, 'baz_plugin')

        self.assertRaises(ValueError,
                          preg.movePluginsUp, IFoo, ('quux_plugin',))

        preg.movePluginsUp(IFoo, ('bar_plugin', 'baz_plugin'))

        idlist = preg.listPluginIds(IFoo)
        self.assertEqual(len(idlist), 3)

        self.assertEqual(idlist[0], 'bar_plugin')
        self.assertEqual(idlist[1], 'baz_plugin')
        self.assertEqual(idlist[2], 'foo_plugin')

        # Moving the top plugin up should not change anything.
        preg.movePluginsUp(IFoo, ('bar_plugin',))
        idlist = preg.listPluginIds(IFoo)
        self.assertEqual(idlist,
                         ('bar_plugin', 'baz_plugin', 'foo_plugin'))

        # Moving the top plugin and another one could change something.
        preg.movePluginsUp(IFoo, ('bar_plugin', 'foo_plugin'))
        idlist = preg.listPluginIds(IFoo)
        self.assertEqual(idlist,
                         ('bar_plugin', 'foo_plugin', 'baz_plugin'))

    def test_movePluginsDown(self):
        from zope.interface import directlyProvides
        plugin_info, IFoo, IBar = self._makePluginInfo()
        parent = DummyFolder()
        foo_plugin = DummyPlugin()
        directlyProvides(foo_plugin, (IFoo,))
        parent._setObject('foo_plugin', foo_plugin)

        bar_plugin = DummyPlugin()
        directlyProvides(bar_plugin, (IFoo,))
        parent._setObject('bar_plugin', bar_plugin)

        baz_plugin = DummyPlugin()
        directlyProvides(baz_plugin, (IFoo,))
        parent._setObject('baz_plugin', baz_plugin)

        preg = self._makeOne(plugin_info).__of__(parent)

        preg.activatePlugin(IFoo, 'foo_plugin')
        preg.activatePlugin(IFoo, 'bar_plugin')
        preg.activatePlugin(IFoo, 'baz_plugin')

        self.assertRaises(ValueError, preg.movePluginsDown,
                          IFoo, ('quux_plugin',))

        preg.movePluginsDown(IFoo, ('foo_plugin', 'bar_plugin'))

        idlist = preg.listPluginIds(IFoo)
        self.assertEqual(len(idlist), 3)

        self.assertEqual(idlist[0], 'baz_plugin')
        self.assertEqual(idlist[1], 'foo_plugin')
        self.assertEqual(idlist[2], 'bar_plugin')

        # Moving the lowest plugin down should not change anything.
        preg.movePluginsDown(IFoo, ('bar_plugin',))
        idlist = preg.listPluginIds(IFoo)
        self.assertEqual(idlist,
                         ('baz_plugin', 'foo_plugin', 'bar_plugin'))

        # Moving the lowest plugin and another one could change something.
        preg.movePluginsDown(IFoo, ('bar_plugin', 'baz_plugin'))
        idlist = preg.listPluginIds(IFoo)
        self.assertEqual(idlist,
                         ('foo_plugin', 'baz_plugin', 'bar_plugin'))

    def test_getAllPlugins(self):
        from zope.interface import directlyProvides
        plugin_info, IFoo, IBar = self._makePluginInfo()
        parent = DummyFolder()
        foo_plugin = DummyPlugin()
        directlyProvides(foo_plugin, (IFoo,))
        parent._setObject('foo_plugin', foo_plugin)

        bar_plugin = DummyPlugin()
        directlyProvides(bar_plugin, (IFoo,))
        parent._setObject('bar_plugin', bar_plugin)

        baz_plugin = DummyPlugin()
        directlyProvides(baz_plugin, (IFoo,))
        parent._setObject('baz_plugin', baz_plugin)

        preg = self._makeOne(plugin_info).__of__(parent)

        first = preg.getAllPlugins('IFoo')

        self.assertEqual(len(first['active']), 0)

        self.assertEqual(len(first['available']), 3)
        self.assertIn('foo_plugin', first['available'])
        self.assertIn('bar_plugin', first['available'])
        self.assertIn('baz_plugin', first['available'])

        preg.activatePlugin(IFoo, 'foo_plugin')

        second = preg.getAllPlugins('IFoo')

        self.assertEqual(len(second['active']), 1)
        self.assertIn('foo_plugin', second['active'])

        self.assertEqual(len(second['available']), 2)
        self.assertNotIn('foo_plugin', second['available'])
        self.assertIn('bar_plugin', second['available'])
        self.assertIn('baz_plugin', second['available'])

        preg.activatePlugin(IFoo, 'bar_plugin')
        preg.activatePlugin(IFoo, 'baz_plugin')

        third = preg.getAllPlugins('IFoo')

        self.assertEqual(len(third['active']), 3)
        self.assertIn('foo_plugin', third['active'])
        self.assertIn('bar_plugin', third['active'])
        self.assertIn('baz_plugin', third['active'])

        self.assertEqual(len(third['available']), 0)

    def test_removePluginById(self):
        from zope.interface import directlyProvides
        plugin_info, IFoo, IBar = self._makePluginInfo()
        parent = DummyFolder()
        foo_plugin = DummyPlugin()
        directlyProvides(foo_plugin, (IFoo, IBar))
        parent._setObject('foo_plugin', foo_plugin)

        bar_plugin = DummyPlugin()
        directlyProvides(bar_plugin, (IFoo,))
        parent._setObject('bar_plugin', bar_plugin)

        baz_plugin = DummyPlugin()
        directlyProvides(baz_plugin, (IBar,))
        parent._setObject('baz_plugin', baz_plugin)

        preg = self._makeOne(plugin_info).__of__(parent)

        preg.activatePlugin(IFoo, 'foo_plugin')
        preg.activatePlugin(IBar, 'foo_plugin')
        preg.activatePlugin(IFoo, 'bar_plugin')
        preg.activatePlugin(IBar, 'baz_plugin')

        preg.removePluginById('foo_plugin')

        idlist = preg.listPluginIds(IFoo)
        self.assertEqual(len(idlist), 1)
        self.assertEqual(idlist[0], 'bar_plugin')

        idlist = preg.listPluginIds(IBar)
        self.assertEqual(len(idlist), 1)
        self.assertEqual(idlist[0], 'baz_plugin')


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(PluginRegistryTests),
    ))



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
""" Classes: PluginRegistry
"""
import logging
import os

import six

from AccessControl import ClassSecurityInfo
from AccessControl.class_init import default__class_init__ as InitializeClass
from AccessControl.Permissions import manage_users as ManageUsers
from Acquisition import aq_inner
from Acquisition import aq_parent
from App.Common import package_home
from App.ImageFile import ImageFile
from OFS.interfaces import IWriteLock
from OFS.SimpleItem import SimpleItem
from Persistence import PersistentMapping
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from zope.interface import implementer

from Products.PluginRegistry.interfaces import IPluginRegistry


try:
    from Products.PluginRegistry.exportimport import PluginRegistryExporter
    from Products.PluginRegistry.exportimport import _updatePluginRegistry
except ImportError:
    _HAS_GENERIC_SETUP = False
else:
    _HAS_GENERIC_SETUP = True

product_dir = package_home(globals())
product_prefix = os.path.split(product_dir)[0]

_wwwdir = os.path.join(product_dir, 'www')

logger = logging.getLogger('PluginRegistry')


@implementer(IPluginRegistry, IWriteLock)
class PluginRegistry(SimpleItem):

    """ Implement IPluginRegistry as an independent, ZMI-manageable object.

    o Each plugin type holds an ordered list of (id, wrapper) tuples.
    """

    security = ClassSecurityInfo()
    meta_type = 'Plugin Registry'
    zmi_icon = 'fas fa-plug'
    _plugins = None

    def __init__(self, plugin_type_info=()):

        if isinstance(plugin_type_info, six.string_types):
            # some tool is passing us our ID.
            raise ValueError('Must pass a sequence of plugin info dicts!')

        self._plugin_types = [x[0] for x in plugin_type_info]
        self._plugin_type_info = PersistentMapping()
        for interface in plugin_type_info:
            self._plugin_type_info[interface[0]] = {
                  'id': interface[1],
                  'title': interface[2],
                  'description': interface[3]}

    #
    #   IPluginRegistry implementation
    #
    @security.protected(ManageUsers)
    def listPluginTypeInfo(self):

        """ See IPluginRegistry.
        """
        result = []

        for ptype in self._plugin_types:

            info = self._plugin_type_info[ptype].copy()
            info['interface'] = ptype
            info['methods'] = list(ptype.names())

            result.append(info)

        return result

    @security.protected(ManageUsers)
    def listPlugins(self, plugin_type):

        """ See IPluginRegistry.
        """
        result = []

        parent = aq_parent(aq_inner(self))

        for plugin_id in self._getPlugins(plugin_type):

            plugin = parent._getOb(plugin_id)
            if not _satisfies(plugin, plugin_type):
                logger.debug('Active plugin %s no longer implements %s' %
                             (plugin_id, plugin_type))
            else:
                result.append((plugin_id, plugin))

        return result

    @security.protected(ManageUsers)
    def getPluginInfo(self, plugin_type):

        """ See IPluginRegistry.
        """
        plugin_type = self._getInterfaceFromName(plugin_type)
        return self._plugin_type_info[plugin_type]

    @security.protected(ManageUsers)
    def listPluginIds(self, plugin_type):

        """ See IPluginRegistry.
        """

        return self._getPlugins(plugin_type)

    @security.protected(ManageUsers)
    def activatePlugin(self, plugin_type, plugin_id):

        """ See IPluginRegistry.
        """
        plugins = list(self._getPlugins(plugin_type))

        if plugin_id in plugins:
            raise KeyError('Duplicate plugin id: {0}'.format(plugin_id))

        parent = aq_parent(aq_inner(self))
        plugin = parent._getOb(plugin_id)

        if not _satisfies(plugin, plugin_type):
            raise ValueError(
                'Plugin does not implement {0}'.format(plugin_type))

        plugins.append(plugin_id)
        self._plugins[plugin_type] = tuple(plugins)

    @security.protected(ManageUsers)
    def deactivatePlugin(self, plugin_type, plugin_id):

        """ See IPluginRegistry.
        """
        plugins = list(self._getPlugins(plugin_type))

        if plugin_id not in plugins:
            raise KeyError('Invalid plugin id: {0}'.format(plugin_id))

        plugins = [x for x in plugins if x != plugin_id]
        self._plugins[plugin_type] = tuple(plugins)

    @security.protected(ManageUsers)
    def movePluginsTop(self, plugin_type, ids_to_move):

        """ See IPluginRegistry.
        """
        ids = list(self._getPlugins(plugin_type))
        indexes = list(map(ids.index, ids_to_move))
        indexes.sort()
        for i1 in indexes:
            ids.insert(0, ids.pop(i1))
        self._plugins[plugin_type] = tuple(ids)

    @security.protected(ManageUsers)
    def movePluginsUp(self, plugin_type, ids_to_move):

        """ See IPluginRegistry.
        """
        ids = list(self._getPlugins(plugin_type))
        count = len(ids)

        indexes = list(map(ids.index, ids_to_move))
        indexes.sort()

        for i1 in indexes:

            if i1 < 0 or i1 >= count:
                raise IndexError(i1)

            i2 = i1 - 1
            if i2 < 0:
                # i1 is already on top
                continue

            ids[i2], ids[i1] = ids[i1], ids[i2]

        self._plugins[plugin_type] = tuple(ids)

    @security.protected(ManageUsers)
    def movePluginsDown(self, plugin_type, ids_to_move):

        """ See IPluginRegistry.
        """
        ids = list(self._getPlugins(plugin_type))
        count = len(ids)

        indexes = list(map(ids.index, ids_to_move))
        indexes.sort()
        indexes.reverse()

        for i1 in indexes:

            if i1 < 0 or i1 >= count:
                raise IndexError(i1)

            i2 = i1 + 1
            if i2 == len(ids):
                # i1 is already on the bottom
                continue

            ids[i2], ids[i1] = ids[i1], ids[i2]

        self._plugins[plugin_type] = tuple(ids)

    #
    #   ZMI
    #
    arrow_right_gif = ImageFile('www/arrow-right.gif', globals())
    arrow_left_gif = ImageFile('www/arrow-left.gif', globals())
    arrow_up_gif = ImageFile('www/arrow-up.gif', globals())
    arrow_down_gif = ImageFile('www/arrow-down.gif', globals())

    @security.protected(ManageUsers)
    def manage_activatePlugins(self, plugin_type, plugin_ids, RESPONSE):
        """ Shim into ZMI.
        """
        interface = self._getInterfaceFromName(plugin_type)
        for id in plugin_ids:
            self.activatePlugin(interface, id)
        RESPONSE.redirect('%s/manage_plugins?plugin_type=%s' %
                          (self.absolute_url(), plugin_type))

    @security.protected(ManageUsers)
    def manage_deactivatePlugins(self, plugin_type, plugin_ids, RESPONSE):
        """ Shim into ZMI.
        """
        interface = self._getInterfaceFromName(plugin_type)
        for id in plugin_ids:
            self.deactivatePlugin(interface, id)

        RESPONSE.redirect('%s/manage_plugins?plugin_type=%s' %
                          (self.absolute_url(), plugin_type))

    @security.protected(ManageUsers)
    def manage_movePluginsUp(self, plugin_type, plugin_ids, RESPONSE):
        """ Shim into ZMI.
        """
        interface = self._getInterfaceFromName(plugin_type)
        self.movePluginsUp(interface, plugin_ids)

        RESPONSE.redirect('%s/manage_plugins?plugin_type=%s' %
                          (self.absolute_url(), plugin_type))

    @security.protected(ManageUsers)
    def manage_movePluginsDown(self, plugin_type, plugin_ids, RESPONSE):
        """ Shim into ZMI.
        """
        interface = self._getInterfaceFromName(plugin_type)
        self.movePluginsDown(interface, plugin_ids)

        RESPONSE.redirect('%s/manage_plugins?plugin_type=%s' %
                          (self.absolute_url(), plugin_type))

    @security.protected(ManageUsers)
    def getAllPlugins(self, plugin_type):

        """ Return a mapping segregating active / available plugins.

        'plugin_type' is the __name__ of the interface.
        """
        interface = self._getInterfaceFromName(plugin_type)

        active = self._getPlugins(interface)
        available = []

        for id, value in aq_parent(aq_inner(self)).objectItems():
            if _satisfies(value, interface):
                if id not in active:
                    available.append(id)

        return {'active': active, 'available': available}

    @security.protected(ManageUsers)
    def removePluginById(self, plugin_id):

        """ Remove a plugin from any plugin types which have it configured.
        """
        for plugin_type in self._plugin_types:

            if plugin_id in self._getPlugins(plugin_type):
                self.deactivatePlugin(plugin_type, plugin_id)

    security.declareProtected(ManageUsers,  # noqa: D001
                              'manage_plugins')
    manage_plugins = PageTemplateFile('plugins', _wwwdir)
    security.declareProtected(ManageUsers,  # noqa: D001
                              'manage_active')
    manage_active = PageTemplateFile('active_plugins', _wwwdir)
    manage_twoLists = PageTemplateFile('two_lists', _wwwdir)

    manage_options = (({'label': 'Plugins', 'action': 'manage_plugins'},
                       {'label': 'Active', 'action': 'manage_active'})
                      + SimpleItem.manage_options)

    if _HAS_GENERIC_SETUP:
        security.declareProtected(ManageUsers,  # noqa: D001
                                  'manage_exportImportForm')
        manage_exportImportForm = PageTemplateFile('export_import', _wwwdir)

        @security.protected(ManageUsers)
        def getConfigAsXML(self):
            """ Return XML representing the registry's configuration.
            """
            pre = PluginRegistryExporter(self).__of__(self)
            return pre.generateXML()

        @security.protected(ManageUsers)
        def manage_exportImport(self, updated_xml, should_purge, RESPONSE):
            """ Parse XML and update the registry.
            """
            # encoding?
            _updatePluginRegistry(self, updated_xml, should_purge)
            RESPONSE.redirect('%s/manage_exportImportForm'
                              '?manage_tabs_message=Registry+updated.' %
                              self.absolute_url())

        @security.protected(ManageUsers)
        def manage_FTPget(self, REQUEST, RESPONSE):
            """
            """
            return self.getConfigAsXML()

        @security.protected(ManageUsers)
        def PUT(self, REQUEST, RESPONSE):
            """
            """
            xml = REQUEST['BODYFILE'].read()
            _updatePluginRegistry(self, xml, True)

        manage_options = (manage_options[:2]
                          + ({'label': 'Export / Import',
                              'action': 'manage_exportImportForm'},)
                          + manage_options[2:])

    #
    #   Helper methods
    #
    @security.private
    def _getPlugins(self, plugin_type):

        if plugin_type not in self._plugin_types:
            raise KeyError(plugin_type)

        if self._plugins is None:
            self._plugins = PersistentMapping()

        return self._plugins.setdefault(plugin_type, ())

    @security.private
    def _getInterfaceFromName(self, plugin_type_name):

        """ Convert the string name to an interface.

        o Raise KeyError if no such interface is known.
        """
        found = [x[0] for x in self._plugin_type_info.items()
                 if x[1]['id'] == plugin_type_name]
        if not found:
            raise KeyError(plugin_type_name)

        if len(found) > 1:
            raise KeyError('Waaa!:  {0}'.format(plugin_type_name))

        return found[0]


InitializeClass(PluginRegistry)


def _satisfies(plugin, iface):
    checker = getattr(iface, 'providedBy', None)
    if checker is None:  # BBB for Zope 2.7?
        checker = iface.isImplementedBy

    return checker(plugin)


def emptyPluginRegistry(ignored):
    """ Return empty registry, for filling from setup profile.
    """
    return PluginRegistry(())



##############################################################################
#
# Copyright (c) 2005 Zope Foundation and Contributors
# Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
""" Unit tests for GenericSetup-based export / import of PluginRegistry.
"""
import unittest


try:
    import Products.GenericSetup  # noqa
except ImportError:
    # No GenericSetup, so no tests
    def test_suite():
        return unittest.TestSuite()
else:
    from Products.GenericSetup.tests.common import BaseRegistryTests
    from Products.GenericSetup.tests.common import DummyExportContext
    from Products.GenericSetup.tests.common import DummyImportContext
    from Products.GenericSetup.utils import _getDottedName
    from zope.component import provideAdapter
    from zope.component.testing import PlacelessSetup
    from zope.interface import Interface
    from zope.interface import directlyProvides
    from zope.traversing.interfaces import ITraversable
    from zope.traversing.interfaces import TraversalError

    class IFoo(Interface):
        pass

    class IBar(Interface):
        pass

    _EMPTY_PLUGINREGISTRY_EXPORT = """\
<?xml version="1.0"?>
<plugin-registry>
</plugin-registry>
"""

    _PLUGIN_TYPE_INFO = (
      (IFoo, 'IFoo', 'foo', 'Some plugin interface'),
      (IBar, 'IBar', 'bar', 'Another plugin interface'))

    _NO_PLUGINS_PLUGINREGISTRY_EXPORT = """\
<?xml version="1.0"?>
<plugin-registry>
 <plugin-type
    id="IFoo"
    interface="%s"
    title="foo"
    description="Some plugin interface">
 </plugin-type>
 <plugin-type
    id="IBar"
    interface="%s"
    title="bar"
    description="Another plugin interface">
 </plugin-type>
</plugin-registry>
""" % (_getDottedName(IFoo), _getDottedName(IBar))

    _NORMAL_PLUGINREGISTRY_EXPORT = """\
<?xml version="1.0"?>
<plugin-registry>
 <plugin-type
    id="IFoo"
    interface="%s"
    title="foo"
    description="Some plugin interface">
  <plugin id="foo_plugin_1" />
  <plugin id="foo_plugin_2" />
 </plugin-type>
 <plugin-type
    id="IBar"
    interface="%s"
    title="bar"
    description="Another plugin interface">
 </plugin-type>
</plugin-registry>
""" % (_getDottedName(IFoo), _getDottedName(IBar))

    class _TestBase(PlacelessSetup, BaseRegistryTests):

        def _initRegistry(self, plugin_type_info=(), plugins={}):
            from OFS.Folder import Folder
            from OFS.SimpleItem import SimpleItem

            from Products.PluginRegistry.PluginRegistry import PluginRegistry

            app = Folder()
            app.getPhysicalPath = lambda: ()
            app.getPhysicalRoot = lambda: app

            app._setObject('foo_plugin_1', SimpleItem())
            app._setObject('foo_plugin_2', SimpleItem())

            registry = PluginRegistry(plugin_type_info)
            registry._plugins = {}  # it is usually lazy

            for plugin_type, registered in plugins.items():
                for obj_id in registered:
                    obj = app._getOb(obj_id)
                    directlyProvides(obj, plugin_type)
                registry._plugins[plugin_type] = registered

            app._setObject('plugin_registry', registry)
            registry = app._getOb('plugin_registry')
            return app, registry

    class PluginRegistryExporterTests(_TestBase):

        def _getTargetClass(self):
            from Products.PluginRegistry.exportimport import \
                PluginRegistryExporter
            return PluginRegistryExporter

        def test_empty(self):
            provideAdapter(AttrItemTraverser, (Interface,), ITraversable)

            app, registry = self._initRegistry()
            exporter = self._makeOne(registry).__of__(registry)
            xml = exporter.generateXML()

            self._compareDOM(xml, _EMPTY_PLUGINREGISTRY_EXPORT)

        def test_normal_no_plugins(self):
            provideAdapter(AttrItemTraverser, (Interface,), ITraversable)

            app, registry = self._initRegistry(
                                    plugin_type_info=_PLUGIN_TYPE_INFO)
            exporter = self._makeOne(registry).__of__(registry)
            xml = exporter.generateXML()

            self._compareDOM(xml, _NO_PLUGINS_PLUGINREGISTRY_EXPORT)

        def test_normal_with_plugins(self):
            provideAdapter(AttrItemTraverser, (Interface,), ITraversable)

            app, registry = self._initRegistry(
                                    plugin_type_info=_PLUGIN_TYPE_INFO,
                                    plugins={IFoo: ('foo_plugin_1',
                                                    'foo_plugin_2')},
                                       )
            exporter = self._makeOne(registry).__of__(registry)
            xml = exporter.generateXML()

            self._compareDOM(xml, _NORMAL_PLUGINREGISTRY_EXPORT)

    class Test_exportPluginRegistry(_TestBase):

        def test_empty(self):
            from Products.PluginRegistry.exportimport import \
                exportPluginRegistry
            provideAdapter(AttrItemTraverser, (Interface,), ITraversable)

            app, registry = self._initRegistry()
            context = DummyExportContext(app)
            exportPluginRegistry(context)

            self.assertEqual(len(context._wrote), 1)
            filename, text, content_type = context._wrote[0]
            self.assertEqual(filename, 'pluginregistry.xml')
            self._compareDOM(text, _EMPTY_PLUGINREGISTRY_EXPORT)
            self.assertEqual(content_type, 'text/xml')

        def test_normal_no_plugins(self):
            from Products.PluginRegistry.exportimport import \
                exportPluginRegistry
            provideAdapter(AttrItemTraverser, (Interface,), ITraversable)

            app, registry = self._initRegistry(
                                    plugin_type_info=_PLUGIN_TYPE_INFO)
            context = DummyExportContext(app)
            exportPluginRegistry(context)

            self.assertEqual(len(context._wrote), 1)
            filename, text, content_type = context._wrote[0]
            self.assertEqual(filename, 'pluginregistry.xml')
            self._compareDOM(text, _NO_PLUGINS_PLUGINREGISTRY_EXPORT)
            self.assertEqual(content_type, 'text/xml')

        def test_normal_with_plugins(self):
            from Products.PluginRegistry.exportimport import \
                exportPluginRegistry
            provideAdapter(AttrItemTraverser, (Interface,), ITraversable)

            app, registry = self._initRegistry(
                                    plugin_type_info=_PLUGIN_TYPE_INFO,
                                    plugins={IFoo: ('foo_plugin_1',
                                                    'foo_plugin_2')})
            context = DummyExportContext(app)
            exportPluginRegistry(context)

            self.assertEqual(len(context._wrote), 1)
            filename, text, content_type = context._wrote[0]
            self.assertEqual(filename, 'pluginregistry.xml')
            self._compareDOM(text, _NORMAL_PLUGINREGISTRY_EXPORT)
            self.assertEqual(content_type, 'text/xml')

    class PluginRegistryImporterTests(_TestBase):

        def _getTargetClass(self):
            from Products.PluginRegistry.exportimport import \
                PluginRegistryImporter
            return PluginRegistryImporter

        def test_parseXML_empty(self):

            app, registry = self._initRegistry()
            importer = self._makeOne(registry).__of__(registry)
            reg_info = importer.parseXML(_EMPTY_PLUGINREGISTRY_EXPORT)

            self.assertEqual(len(reg_info['plugin_types']), 0)

        def test_parseXML_normal_no_plugins(self):

            app, registry = self._initRegistry()
            importer = self._makeOne(registry).__of__(registry)
            reg_info = importer.parseXML(_NO_PLUGINS_PLUGINREGISTRY_EXPORT)

            self.assertEqual(len(reg_info['plugin_types']), 2)

            info = reg_info['plugin_types'][0]
            self.assertEqual(info['id'], 'IFoo')
            self.assertEqual(info['interface'], _getDottedName(IFoo))
            self.assertEqual(info['title'], 'foo')
            self.assertEqual(info['description'], 'Some plugin interface')
            self.assertEqual(len(info['plugins']), 0)

            info = reg_info['plugin_types'][1]
            self.assertEqual(info['id'], 'IBar')
            self.assertEqual(info['interface'], _getDottedName(IBar))
            self.assertEqual(info['title'], 'bar')
            self.assertEqual(info['description'], 'Another plugin interface')
            self.assertEqual(len(info['plugins']), 0)

        def test_parseXML_normal_with_plugins(self):

            app, registry = self._initRegistry()
            importer = self._makeOne(registry).__of__(registry)
            reg_info = importer.parseXML(_NORMAL_PLUGINREGISTRY_EXPORT)

            self.assertEqual(len(reg_info['plugin_types']), 2)

            info = reg_info['plugin_types'][0]
            self.assertEqual(info['id'], 'IFoo')
            self.assertEqual(info['interface'], _getDottedName(IFoo))
            self.assertEqual(info['title'], 'foo')
            self.assertEqual(info['description'], 'Some plugin interface')
            plugins = info['plugins']
            self.assertEqual(len(plugins), 2)
            self.assertEqual(plugins[0]['id'], 'foo_plugin_1')
            self.assertEqual(plugins[1]['id'], 'foo_plugin_2')

            info = reg_info['plugin_types'][1]
            self.assertEqual(info['id'], 'IBar')
            self.assertEqual(info['interface'], _getDottedName(IBar))
            self.assertEqual(info['title'], 'bar')
            self.assertEqual(info['description'], 'Another plugin interface')
            self.assertEqual(len(info['plugins']), 0)

    class Test_importPluginRegistry(_TestBase):

        def test_empty_default_purge(self):
            from Products.PluginRegistry.exportimport import \
                importPluginRegistry

            app, registry = self._initRegistry(
                                    plugin_type_info=_PLUGIN_TYPE_INFO,
                                    plugins={IFoo: ('foo_plugin_1',
                                                    'foo_plugin_2')})

            self.assertEqual(len(registry.listPluginTypeInfo()), 2)
            self.assertEqual(len(registry.listPlugins(IFoo)), 2)
            self.assertEqual(len(registry.listPlugins(IBar)), 0)

            context = DummyImportContext(app)
            context._files['pluginregistry.xml'] = _EMPTY_PLUGINREGISTRY_EXPORT

            importPluginRegistry(context)

            self.assertEqual(len(registry.listPluginTypeInfo()), 0)
            self.assertRaises(KeyError, registry.listPlugins, IFoo)
            self.assertRaises(KeyError, registry.listPlugins, IBar)

        def test_empty_explicit_purge(self):
            from Products.PluginRegistry.exportimport import \
                importPluginRegistry

            app, registry = self._initRegistry(
                                    plugin_type_info=_PLUGIN_TYPE_INFO,
                                    plugins={IFoo: ('foo_plugin_1',
                                                    'foo_plugin_2')})

            self.assertEqual(len(registry.listPluginTypeInfo()), 2)
            self.assertEqual(len(registry.listPlugins(IFoo)), 2)
            self.assertEqual(len(registry.listPlugins(IBar)), 0)

            ctx = DummyImportContext(app, True)
            ctx._files['pluginregistry.xml'] = _EMPTY_PLUGINREGISTRY_EXPORT

            importPluginRegistry(ctx)

            self.assertEqual(len(registry.listPluginTypeInfo()), 0)
            self.assertRaises(KeyError, registry.listPlugins, IFoo)
            self.assertRaises(KeyError, registry.listPlugins, IBar)

        def test_empty_skip_purge(self):
            from Products.PluginRegistry.exportimport import \
                importPluginRegistry

            app, registry = self._initRegistry(
                                    plugin_type_info=_PLUGIN_TYPE_INFO,
                                    plugins={IFoo: ('foo_plugin_1',
                                                    'foo_plugin_2')})

            self.assertEqual(len(registry.listPluginTypeInfo()), 2)
            self.assertEqual(len(registry.listPlugins(IFoo)), 2)
            self.assertEqual(len(registry.listPlugins(IBar)), 0)

            ctx = DummyImportContext(app, False)
            ctx._files['pluginregistry.xml'] = _EMPTY_PLUGINREGISTRY_EXPORT

            importPluginRegistry(ctx)

            self.assertEqual(len(registry.listPluginTypeInfo()), 2)
            self.assertEqual(len(registry.listPlugins(IFoo)), 2)
            self.assertEqual(len(registry.listPlugins(IBar)), 0)

        def test_normal_no_plugins(self):
            from Products.PluginRegistry.exportimport import \
                importPluginRegistry

            app, registry = self._initRegistry(
                              plugins={IFoo: ('foo_plugin_1', 'foo_plugin_2')})

            self.assertEqual(len(registry.listPluginTypeInfo()), 0)
            self.assertRaises(KeyError, registry.listPlugins, IFoo)
            self.assertRaises(KeyError, registry.listPlugins, IBar)

            ctx = DummyImportContext(app, False)
            xml_data = _NO_PLUGINS_PLUGINREGISTRY_EXPORT
            ctx._files['pluginregistry.xml'] = xml_data

            importPluginRegistry(ctx)

            self.assertEqual(len(registry.listPluginTypeInfo()), 2)

            info = registry.listPluginTypeInfo()[0]
            self.assertEqual(info['id'], 'IFoo')
            self.assertEqual(info['title'], 'foo')
            self.assertEqual(info['description'], 'Some plugin interface')

            info = registry.listPluginTypeInfo()[1]
            self.assertEqual(info['id'], 'IBar')
            self.assertEqual(info['title'], 'bar')
            self.assertEqual(info['description'], 'Another plugin interface')

            self.assertEqual(len(registry.listPlugins(IFoo)), 0)
            self.assertEqual(len(registry.listPlugins(IBar)), 0)

        def test_normal_with_plugins(self):
            from Products.PluginRegistry.exportimport import \
                importPluginRegistry

            app, registry = self._initRegistry(
                              plugins={IFoo: ('foo_plugin_1', 'foo_plugin_2')})

            self.assertEqual(len(registry.listPluginTypeInfo()), 0)
            self.assertRaises(KeyError, registry.listPlugins, IFoo)
            self.assertRaises(KeyError, registry.listPlugins, IBar)

            ctx = DummyImportContext(app, False)
            ctx._files['pluginregistry.xml'] = _NORMAL_PLUGINREGISTRY_EXPORT

            importPluginRegistry(ctx)

            self.assertEqual(len(registry.listPluginTypeInfo()), 2)

            info = registry.listPluginTypeInfo()[0]
            self.assertEqual(info['id'], 'IFoo')
            self.assertEqual(info['title'], 'foo')
            self.assertEqual(info['description'], 'Some plugin interface')

            info = registry.listPluginTypeInfo()[1]
            self.assertEqual(info['id'], 'IBar')
            self.assertEqual(info['title'], 'bar')
            self.assertEqual(info['description'], 'Another plugin interface')

            self.assertEqual(len(registry.listPlugins(IFoo)), 2)
            plugins = registry.listPlugins(IFoo)
            self.assertEqual(plugins[0][0], 'foo_plugin_1')
            self.assertEqual(plugins[0][1], app._getOb('foo_plugin_1'))
            self.assertEqual(plugins[1][0], 'foo_plugin_2')
            self.assertEqual(plugins[1][1], app._getOb('foo_plugin_2'))

            self.assertEqual(len(registry.listPlugins(IBar)), 0)

        def test_normal_with_plugins_skip_duplicates(self):
            # See http://www.zope.org/Collectors/PAS/52
            from Products.PluginRegistry.exportimport import \
                importPluginRegistry

            app, registry = self._initRegistry()

            self.assertEqual(len(registry.listPluginTypeInfo()), 0)
            self.assertRaises(KeyError, registry.listPlugins, IFoo)
            self.assertRaises(KeyError, registry.listPlugins, IBar)

            ctx = DummyImportContext(app, False)
            ctx._files['pluginregistry.xml'] = _NORMAL_PLUGINREGISTRY_EXPORT

            importPluginRegistry(ctx)
            importPluginRegistry(ctx)  # twice should not duplicate

            self.assertEqual(len(registry.listPluginTypeInfo()), 2)

    class AttrItemTraverser:
        _marker = object()

        def __init__(self, context):
            self.context = context

        def traverse(self, name, furtherPath):
            result = getattr(self.context, name, self._marker)
            if result is self._marker:
                try:
                    result = self.context.get(name, self._marker)
                except AttributeError:
                    pass
            if result is self._marker:
                raise TraversalError(name)
            return result

    def test_suite():
        return unittest.TestSuite((
            unittest.makeSuite(PluginRegistryExporterTests),
            unittest.makeSuite(PluginRegistryImporterTests),
            unittest.makeSuite(Test_exportPluginRegistry),
            unittest.makeSuite(Test_importPluginRegistry)))

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
""" PluginRegistry product initialization.
"""


def initialize(context):
    from . import PluginRegistry
    context.registerClass(PluginRegistry.PluginRegistry,
                          constructors=[('Dummy', lambda: None)],
                          visibility=None,
                          icon='www/PluginRegistry.png')



from setuptools import find_packages
from setuptools import setup


NAME = 'PluginRegistry'

with open('README.rst') as f:
    README = f.read()

with open('CHANGES.rst') as f:
    CHANGES = f.read()

setup(
    name='Products.PluginRegistry',
    version='1.10',
    description='Configure application plugins based on interfaces',
    long_description=README + '\n\n' + CHANGES,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Plone',
        'Framework :: Plone :: 5.2',
        'Framework :: Plone :: Core',
        'Framework :: Zope',
        'Framework :: Zope :: 4',
        'Framework :: Zope :: 5',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development',
        'Topic :: System :: Archiving :: Packaging',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
    keywords='web application server zope zope2',
    author='Zope Foundation and Contributors',
    author_email='zope-cmf@lists.zope.org',
    url='https://github.com/zopefoundation/Products.PluginRegistry',
    project_urls={
        'Issue Tracker': ('https://github.com/zopefoundation/Products.'
                          'PluginRegistry/issues'),
        'Sources': 'https://github.com/zopefoundation/Products.PluginRegistry',
    },
    license='ZPL 2.1',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    namespace_packages=['Products'],
    zip_safe=False,
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
    install_requires=[
        'setuptools',
        'six',
        'Zope >= 4.0b4',
        'Products.GenericSetup >= 2.0b1',
    ],
    entry_points="""
    [zope2.initialize]
    Products.PluginRegistry = Products.PluginRegistry:initialize
    """,
)
